% Cluster genes separately at each pool, where each pool is a set of time points,
% using a dependent generative model for their expression levels and gene
% ontology (GO) terms.
%
% This routine does multimodal clustering of gene expression profiles and GO
% tags associated with them. A generative model is used for clustering.
% According to the model, a gene g's expression vector '{e_gs}' for the samples 
% 's' corresponding to a state or pool 'p' and its associated GO tags 'O_g' are
% assumed to be generated from a state or pool specific cluster 'c_p'. The GO
% tags are generated through the expression levels. Further each tag gets
% generated at any one of the possible states within the training data. The
% Gaussian parameters for the time points corresponding to a state are tied,
% i.e. constrained to be the same. The parameters are estimated using EM.
%
% The input parameter 'expr_data_file' is a data file containing a matrix of
% expression values with each row corresponding to a gene and each column
% corresponding to a tissue sample or time-point.
%
% 'go_tags_file' is a file containing the matrix of integer IDs of GO tags for
% the genes. Each row corresponds to a gene and each column to one of its GO
% tags. Since different genes have different numbers of annotations, the number
% of columns in the matrix is equal to the maximum number of annotations for any
% gene. The columns for each gene are first filled with positive integers
% corresponding to its GO tag IDs followed by -1's if the gene does not have the
% maximum number of annotations. The first column is an integer ID for the gene
% itself. 
%
% 'n_clusters' is a vector specifying the number of clusters to group the data
% for each state or pool. Its length is equal to the number of
% states or pools. This allows for different states or pools to
% have different numbers of clusters.
%
% 'n_iters' is the number of EM iterations.
%
% 'do_weight_GO_tags' specifies whether the weights P(g|o) have to be considered
% or not. When it is set to 0, P(g|o) is 1.0 everywhere, otherwise it is a
% uniform distribution over all genes 'g' that are annotated with the term 'o'.
%
% 'do_count_GO_tags_once' takes a mean of the contributions from the GO
% tag probabilities conditioned on the pool specific clusters across all the
% time points. This is equivalent to counting GO tags once for every gene
% regardless of the number of time points for which its expression level is
% measured in the training data.
%   
% 'do_arithmetic_avg' makes sense only when the 'do_count_GO_tags_once' option
% is enabled. It leads to arithmetic average being taken instead of geometric
% average to count GO tags only once across all the time points. This option
% does the true dependent generative model. It makes more sense to enable this
% option along with 'do_count_GO_tags_once'.
%
% 'output_mat_file' is the name of a .mat file into which model parameters are
% written. These include the means and variances of the Gaussian mixtures of
% each sample or time-point and the estimated joint distributions P(o, c_s) over
% ontology terms and clusters.
%
% 'forced_state_switch_pts' is a vector of time points at which the forced
% switching from one state to the next happen. The number of such time points
% should be one less than the number of states or pools. If the switch time
% points are not specified in increasing order, then the vector is sorted before
% using the values.

function [] = cluster_with_em_GO_tags_state_specific_dependent_tied_params(expr_data_file, ...
                                                                           go_tags_file, ...
                                                                           n_states, ...
                                                                           n_clusters, ...
                                                                           n_iters, ...
                                                                           do_weight_GO_tags, ...
                                                                           do_count_GO_tags_once, ...
                                                                           do_share_clusters, ...
                                                                           do_arithmetic_avg, ...
                                                                           output_mat_file, ...
                                                                           forced_state_switch_pts)


% Load the data from files.
input_gene_exprs   = load(expr_data_file);
input_go_tags_data = load(go_tags_file);

% To allow variation in the size of training set.
gene_exprs   = input_gene_exprs;
go_tags_data = input_go_tags_data;

% Useful variables.
n_samples                 = size(gene_exprs, 2);
total_num_feature_vectors = size(gene_exprs, 1);
num_exprs_per_gene        = 1;
n_genes                   = total_num_feature_vectors/num_exprs_per_gene;

n_fields          = size(go_tags_data, 2);
num_tags_per_gene = n_fields - 1;

% Reshape the feature data into a 3-dimensional matrix such that:
% 1. each plane corresponds to an gene.
% 2. in each plane, rows correspond to regions and columns correspond to
% feature dimensions.
feature = zeros(num_exprs_per_gene, n_samples, n_genes);
for i = 1:n_genes
    feature(:, :, i) = gene_exprs((num_exprs_per_gene*(i - 1) + 1):(num_exprs_per_gene*i), :);
end

word        = zeros(num_tags_per_gene, 1, n_genes);
n_go_tags_g = zeros(n_genes, 1);
for i = 1:n_genes
    word(:, 1, i)  = (go_tags_data(i, 2:n_fields))';
    n_go_tags_g(i) = length(find(word(:, 1, i) > 0));
end

n_go_tags = max(word(:))
M         = max(n_clusters);

%%%%%%%%%%%% Debugging%%%%%%%%%%%
%feature(:, :, i)
%word(:, :, i)
%%%%%%%%%%%% Debugging%%%%%%%%%%%

if ( ~isempty(forced_state_switch_pts) )
    
    sorted_state_switch_pts = sort(forced_state_switch_pts);
    n_state_switches        = length(sorted_state_switch_pts);

    if ( (sorted_state_switch_pts(1) < 2) | ...
         (sorted_state_switch_pts(n_state_switches) > n_samples) | ...
         (n_state_switches < (n_states-1)) | ...
         (n_state_switches > (n_states-1)) )
       
        disp('The vector specifying state switch points is invalid!');
        return;

    end

    state_samples   = {};
    prev_switch_pt  = 1;
    curr_switch_idx = 1;
    curr_s          = 1;    
    while (curr_switch_idx <= n_state_switches)
        next_switch_pt        = sorted_state_switch_pts(curr_switch_idx);
        state_samples{curr_s} = prev_switch_pt:(next_switch_pt-1);
        state_n_samples{curr_s} = length(state_samples{curr_s});

        prev_switch_pt  = next_switch_pt;
        curr_switch_idx = curr_switch_idx + 1;
        curr_s          = curr_s + 1;
    end
    state_samples{curr_s}   = prev_switch_pt:n_samples;
    state_n_samples{curr_s} = length(state_samples{curr_s});

    
else
    disp('The vector specifying state switch points is empty!');
    return;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Model parameters.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%M = 100; % Number of concepts.

P_c = zeros(M, n_states); % Concept prior.

% Generative distributions of concepts.
% Word generator:
P_w_given_c = zeros(M, n_go_tags, n_states);

% Region generator:
u_c     = zeros(M, n_states);
v_c     = zeros(M, n_states);
stdev_c = zeros(M, n_states);

% Probabilities linking words and regions to concepts:
P_c_given_x_i = rand(num_exprs_per_gene, M, n_genes, n_states).^5;
P_c_given_w_i = zeros(num_tags_per_gene, M, n_genes, n_states);

for s = 1:n_states

    % Initialize only those elements that correspond to a valid vocabulary
    % word for every gene:
    for k = 1:num_tags_per_gene
        for j = 1:1
            for i = 1:n_genes
                if (word(k, j, i) > 0)
                    P_c_given_w_i(k, :, i, s) = rand(1, M).^5;
                end
            end
        end
    end

    % Normalize the initialized arrays.
    for i = 1:n_genes
        
        % Normalize P(c|x, i).
        for j = 1:num_exprs_per_gene
            P_c_given_x_i(j, :, i, s) = P_c_given_x_i(j, :, i, s)./sum(P_c_given_x_i(j, :, i, s));
        end
        
        % Normalize P(c|w, i).
        for j = 1:num_tags_per_gene
            % Consider only valid words.
            if(word(j, 1, i) > 0)
                P_c_given_w_i(j, :, i, s) = P_c_given_w_i(j, :, i, s)./sum(P_c_given_w_i(j, :, i, s));
            end
        end
        
    end

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% EM algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Useful data structures.
sum_P_c_given_x_i = zeros(M, n_states);
P_c_given_X_i     = zeros(M, 1, n_genes, n_states);
P_c_given_X_g     = zeros(M, n_genes, n_states);
P_c_given_w_g     = zeros(num_tags_per_gene, M, n_genes, n_states);

log_prod_stdev_c = zeros(M, n_states); 
log_P_x_c        = zeros(M, num_exprs_per_gene, n_genes, n_states);
log_P_tilda_x_c  = zeros(M, num_exprs_per_gene, n_genes, n_states);
log_P_c          = zeros(M, n_states);
M_x              = zeros(num_exprs_per_gene, 1, n_genes, n_states);         

log_P_w_given_X = zeros(num_tags_per_gene, 1, n_genes, n_states); 
log_P_tilda_x   = zeros(num_exprs_per_gene, 1, n_genes, n_states);
log_likelihood  = zeros(n_iters, 1);

for iter = 1:n_iters
    
    iteration = iter
    
    log_likelihood(iter) = 0;

    for s = 1:n_states
        
        %%%%%%%%%%%%
        % M-step.
        %%%%%%%%%%%%
        
        % Compute P(c).
        for c = 1:M
            P_c(c, s) = 0;
            for i = 1:n_genes
            
                % Consider only the valid words.
                for j = 1:num_tags_per_gene
                    if (word(j, 1, i) > 0)                        
                        if (do_arithmetic_avg)
                            P_c(c, s) = P_c(c, s) + (P_c_given_w_i(j, c, i, s)/n_states);
                        else
                            P_c(c, s) = P_c(c, s) + P_c_given_w_i(j, c, i, s);
                        end
                    end
                end
                
                P_c(c, s) = P_c(c, s) + P_c_given_x_i(:, c, i, s);
            end
        end
        
        % Compute P(w|c).
        for c = 1:M
            %for j = 1:n_go_tags % Too slow.
                P_w_given_c(c, :, s) = 0;
                for i = 1:n_genes
                    
                    % Too slow: Consider only the valid words and the word in question. 
                    % Faster: Consider only the valid words. 
                    for k = 1:num_tags_per_gene
                        if (word(k, 1, i) > 0)
                            j = word(k, 1, i);
                            if (do_arithmetic_avg)
                                P_w_given_c(c, j, s) = P_w_given_c(c, j, s) + (P_c_given_w_i(k, c, i, s)/n_states);
                            else
                                P_w_given_c(c, j, s) = P_w_given_c(c, j, s) + P_c_given_w_i(k, c, i, s);
                            end
                        end
                    end
                    
                end
            %end % Too slow.
        end
        
        % Compute u_c.
        for c = 1:M
            u_c(c, s)              = 0;
            sum_P_c_given_x_i(c,s) = 0;
            for i = 1:n_genes
                for j = 1:num_exprs_per_gene
                    u_c(c, s)               = u_c(c, s) + (sum(feature(j, state_samples{s}, i))*P_c_given_x_i(j, c, i, s)/state_n_samples{s});
                    sum_P_c_given_x_i(c, s) = sum_P_c_given_x_i(c, s) + P_c_given_x_i(j, c, i, s);
                end
            end
            
            u_c(c, s) = u_c(c, s)./sum_P_c_given_x_i(c, s);
        end
        
        % Compute v_c.
        % Note that sum_P_c_given_x_i values are already available.
        for c = 1:M
            v_c(c, s) = 0;
            for i = 1:n_genes
                for j = 1:num_exprs_per_gene
                    v_c(c, s) = v_c(c, s) + (sum(((feature(j, state_samples{s}, i) - u_c(c, s)).*(feature(j, state_samples{s}, i) - u_c(c, s))))*P_c_given_x_i(j, c, i, s)/state_n_samples{s});
                end
            end
            
            v_c(c, s) = v_c(c, s)./sum_P_c_given_x_i(c, s);
        end
        
        % Normalize the probabilities so that they sum to 1.
        P_c(:, s) = P_c(:, s)./sum(P_c(:, s));
        for c = 1:M
            P_w_given_c(c, :, s) = P_w_given_c(c, :, s)./sum(P_w_given_c(c, :, s));
        end
        
        % Compute P_c in log space. Useful in E-step and LL calculations.
        log_P_c(:, s) = safe_log(P_c(:, s));
        
        %%%%%%%%%%%%
        % E-step.
        %%%%%%%%%%%%
        
        % Useful data structure.
        stdev_c(:, s) = sqrt(v_c(:, s));
        
            % Form log_P_x_c.
            for c = 1:M
                log_product_stdev_c(c, s) = state_n_samples{s}*safe_log(prod(stdev_c(c, s)));
                for i = 1:n_genes
                    
                    for j = 1:num_exprs_per_gene
                        log_P_x_c(c, j, i, s) = (-0.5*sum(((feature(j, state_samples{s}, i) - u_c(c, s)).*(feature(j, state_samples{s}, i) - u_c(c, s)))./v_c(c, s)));
                        log_P_x_c(c, j, i, s) = log_P_x_c(c, j, i, s) - log_product_stdev_c(c, s);
                        log_P_x_c(c, j, i, s) = log_P_c(c, s) + log_P_x_c(c, j, i, s);
                    end
                end
            end
            
            % Form log_P_tilda_x_c:
            % 1. Compute M_x.
            % 2. Subtract M_x from log_P_x_c.
            for i = 1:n_genes
                for j = 1:num_exprs_per_gene
                    M_x(j, 1, i, s) = max(log_P_x_c(:, j, i, s));
                end
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %debug_log_P_x_c = log_P_x_c(:, :, i)
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                
                for c = 1:M
                    log_P_tilda_x_c(c, :, i, s) = log_P_x_c(c, :, i, s) - M_x(:, 1, i, s)';
                    
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    %debug_M_x = M_x(:, 1, i)'
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                end
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %for j = 1:num_exprs_per_gene
                    %debug_log_P_tilda_x_c     = log_P_tilda_x_c(:, j, i)
                    %debug_max_log_P_tilda_x_c = max(log_P_tilda_x_c(:, j, i))
                    %pause;
                %end
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
            end
            
            % Compute P(c|x, i), P(c|X, i) and P(c|w, i).
            for i = 1:n_genes
                
                % Compute P(c|x, i).
                for c = 1:M
                     P_c_given_x_i(:, c, i, s) = (exp(log_P_tilda_x_c(c, :, i, s)))';
                end
                
                % Normalize P(c|x, i).
                for j = 1:num_exprs_per_gene
                    P_c_given_x_i(j, :, i, s) = P_c_given_x_i(j, :, i, s)./sum(P_c_given_x_i(j, :, i, s));
                    
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    %debug_P_c_given_x_i     = P_c_given_x_i(j, :, i)
                    %debug_sum_P_c_given_x_i = sum(P_c_given_x_i(j, :, i))
                    %%%%%%%%%%%% Debugging%%%%%%%%%%%
                end
                
                % Compute P(c|X, i).
                for c = 1:M
                    P_c_given_X_i(c, 1, i, s) = sum(P_c_given_x_i(:, c, i, s));
                end
                
                % Normalize P(c|X, i).
                P_c_given_X_i(:, 1, i, s) = P_c_given_X_i(:, 1, i, s)./sum(P_c_given_X_i(:, 1, i, s));
                
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                %debug_P_c_given_X_i     = P_c_given_X_i(:, 1, i)
                %debug_sum_P_c_given_X_i = sum(P_c_given_X_i(:, 1, i))
                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    
                % Compute P(c|w, i).
                for j = 1:num_tags_per_gene
                        % Consider only valid words.
                        if(word(j, 1, i) > 0)
                            P_c_given_w_i(j, :, i, s) = P_w_given_c(:, word(j, 1, i), s).*P_c_given_X_i(:, 1, i, s);

                            if (do_count_GO_tags_once)
                                if (do_arithmetic_avg)
                                    log_P_w_given_X(j, 1, i, s) = safe_log(sum(P_c_given_w_i(j, :, i, s))/n_states);
                                else
                                    log_P_w_given_X(j, 1, i, s) = safe_log((sum(P_c_given_w_i(j, :, i, s)))^(1/n_states));
                                end
                            else
                                log_P_w_given_X(j, 1, i, s) = safe_log(sum(P_c_given_w_i(j, :, i, s)));
                            end

                        end
                end
                
                % Normalize P(c|w, i).
                for j = 1:num_tags_per_gene
                    % Consider only valid words.
                    if(word(j, 1, i) > 0)
                        P_c_given_w_i(j, :, i, s) = P_c_given_w_i(j, :, i, s)./sum(P_c_given_w_i(j, :, i, s));
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_P_c_given_w_i     = P_c_given_w_i(j, :, i)
                        %debug_sum_P_c_given_w_i = sum(P_c_given_w_i(j, :, i))
                        %pause;
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    end
                end
                
            end
                
                                
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Log likelihood computation.
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            for i = 1:n_genes
                
                for j = 1:num_exprs_per_gene
                    log_P_tilda_x(j, 1, i, s) = safe_log(sum(exp(log_P_tilda_x_c(:, j, i, s)))); 
                    
                    log_likelihood(iter) = log_likelihood(iter) + M_x(j, 1, i, s);
                    log_likelihood(iter) = log_likelihood(iter) + log_P_tilda_x(j, 1, i, s);
                end
                
                if (~do_arithmetic_avg)
                    for j = 1:num_tags_per_gene
                        % Consider only valid words.
                        if (word(j, 1, i) > 0)                                                
                            log_likelihood(iter) = log_likelihood(iter) + (log_P_w_given_X(j, 1, i, s)/n_go_tags_g(i));
                        end
                    end
                end
            end

    end
    
    if (do_arithmetic_avg)
        for i = 1:n_genes
            for j = 1:num_tags_per_gene
                sum_over_s = 0;
                for s = 1:n_states
                    if (word(j, 1, i) > 0)
                        sum_over_s = sum_over_s + exp(log_P_w_given_X(j, 1, i, s));
                    end
                end
                log_likelihood(iter) = log_likelihood(iter) + (safe_log(sum_over_s)/n_go_tags_g(i));
            end
        end
    end
    
    %%%%%%%%%%%% Debugging%%%%%%%%%%%
    %P_c
    %P_w_given_c
    %u_c
    %v_c
    log_likelihood
    %%%%%%%%%%%% Debugging%%%%%%%%%%%

end

if ( ~isempty(forced_state_switch_pts) )
    
    curr_switch_idx = 1;
    curr_s          = 1;    
    while (curr_switch_idx <= n_state_switches)

        P_c_given_X_g(:, :, curr_s)    = 0;
        P_c_given_w_g(:, :, :, curr_s) = 0;

        % Average the sample specific cluster posteriors into state specific posteriors.
        for i = 1:n_genes
                        
            % Process P(c|x, i).
            for j = 1:num_exprs_per_gene
                for c = 1:M
                    P_c_given_X_g(c, i, curr_s) = P_c_given_X_g(c, i, curr_s) + P_c_given_x_i(j, c, i, curr_s);
                end
            end
            
            % Process P(c|w, i).
            for j = 1:num_tags_per_gene
                % Consider only valid words.
                if(word(j, 1, i) > 0)
                    for c = 1:M
                        P_c_given_w_g(j, c, i, curr_s) = P_c_given_w_g(j, c, i, curr_s) + P_c_given_w_i(j, c, i, curr_s);
                    end
                    P_c_given_w_g(j, :, i, curr_s) = P_c_given_w_g(j, :, i, curr_s) ./ sum(P_c_given_w_g(j, :, i, curr_s));
                end
            end
        
            % Normalize.
            P_c_given_X_g(:, i, curr_s) = P_c_given_X_g(:, i, curr_s) ./ sum(P_c_given_X_g(:, i, curr_s));

        end
        
        curr_switch_idx = curr_switch_idx + 1;
        curr_s          = curr_s + 1;

    end
    % Do it for the last state.
    P_c_given_X_g(:, :, curr_s)    = 0;
    P_c_given_w_g(:, :, :, curr_s) = 0;

    % Average the sample specific cluster posteriors into state specific posteriors.
    for i = 1:n_genes
                    
        % Process P(c|x, i).
        for j = 1:num_exprs_per_gene
            for c = 1:M
                P_c_given_X_g(c, i, curr_s) = P_c_given_X_g(c, i, curr_s) + P_c_given_x_i(j, c, i, curr_s);
            end
        end
        
        % Process P(c|w, i).
        for j = 1:num_tags_per_gene
            % Consider only valid words.
            if(word(j, 1, i) > 0)
                for c = 1:M
                    P_c_given_w_g(j, c, i, curr_s) = P_c_given_w_g(j, c, i, curr_s) + P_c_given_w_i(j, c, i, curr_s);
                end
                P_c_given_w_g(j, :, i, curr_s) = P_c_given_w_g(j, :, i, curr_s) ./ sum(P_c_given_w_g(j, :, i, curr_s));
            end
        end
    
        % Normalize.
        P_c_given_X_g(:, i, curr_s) = P_c_given_X_g(:, i, curr_s) ./ sum(P_c_given_X_g(:, i, curr_s));

    end

end
% Save the model parameters to a .mat file so that it can be loaded later.
save(output_mat_file, 'state_samples', ...
                      'n_clusters', ...
                      'P_c', ...
                      'P_c_given_X_g', ...
                      'P_c_given_w_g', ...
                      'P_w_given_c', ...
                      'u_c', ...
                      'v_c');
