% Get Gene Ontology (GO) ancestors for a previously annotated file.
%
% This routine appends the GO IDs in a previously annotated file with all their
% ancestors in the GO ontology tree.
%
% The parameter 'go_annots_inp_file' is the name of the file containing the
% annotations (without ancestors) for a set of genes. It is assumed to be a two
% column file with the first column containing gene IDs and the second column
% containing GO IDs for the corresponding genes with ';' as the separator
% between multiple GO IDs.
%
% The parameter 'go_annots_out_file' specifies the name of the output file into
% which the original GO IDs go along with the IDs of their ancestors appended.
% It has the same format as the input file. 
%
% 'root_id' is the numeric id of the root GO term in the particular GO graph
% being used. It would be one of the following:
% Biological Process(GO:0008150): 8150 
% Molecular function(GO:0003674): 3674
% Cellular Component(GO:0005575): 5575
% This id is used to force the root as the ancestor for those genes with
% obsolete ids.
%
% WARNINGS: 
%
% 1. It runs into problems when the GO tags in the input file are not
% numerical IDs. This is usually not expected but was seen in one of the
% annotation files (molecular function) for the NIA Mouse 15K data. The problem lies in
% how a string ID with spaces in it gets split while reading with the textread()
% function. Need a robust way to deal with this.
%
% 2. It does not explicitly deal with a header line due to textread(). So the
% header might need to be adjusted in the output file. This can be fixed with a
% bit more effort (TBD!).

function [] = get_GO_ancestors(go_annots_inp_file, ...
                               go_annots_out_file, ...
                               root_id)

[gene_ids, go_ids] = textread(go_annots_inp_file, '%s%s');

num_genes = length(gene_ids);

GO = geneont('live', true);

[fid, msg] = fopen(go_annots_out_file, 'w');

if (fid == -1)
    disp('Cannot open file\n');
    disp(msg);
    return;
end

for i = 1:num_genes    
    % Debugging.
    %i
    fprintf(fid, '%s\t%s', gene_ids{i}, go_ids{i});
    k = strfind(go_ids{i}, 'GO:');

    num_go_ids       = length(k);
    go_ids_str       = char(go_ids{i});
    go_ids_w_anc_str = go_ids_str;

    for j = 1:num_go_ids
        if (~isletter(str2num(go_ids_str(k(j)+3))))
            go_id   = str2num(go_ids_str((k(j)+3):(k(j)+9)));

            anc     = getancestors(GO, go_id);
            % Append root id.
            anc     = [anc; root_id];
            num_anc = length(anc);

            for l = 1:num_anc
                anc_id_str = sprintf('GO:%07d', anc(l));

                % Do not repeat.
                m = strfind(go_ids_w_anc_str, anc_id_str);
                if (isempty(m))
                    go_ids_w_anc_str = strcat(go_ids_w_anc_str, anc_id_str);
                    fprintf(fid, ';%s', anc_id_str);
                end
            end
        end
    end
    
    fprintf(fid, '\n');
end

fclose(fid);
