% Parse GO annotation files provided by the GO consortium.
%
% This routine parses the GO annotation files downloaded from the GO Consortium
% website. The Matlab function goannotread() does most of the parsing. The
% parsed output from the Matlab function is split into tags corresponding to the
% Biological Process (BP), Molecular Function (MF) and Cellular Component (CC)
% sub-ontologies and written to the corresponding output files.
%
% The parameter 'go_consortium_annots_file' specifies the name of the annotation
% file downloaded from the GO Consortium website.
%
% The parameters 'output_xx_annots_file' specify the names of the output files
% corresponding to the three sub-ontologies. Each output file has two columns of
% data. The first column lists the gene symbols and the second column the GO
% tags corresponding to the gene in the first column. In case of multiple GO
% tags for a gene, they  are separated by semicolons. If a gene is represented
% by multiple symbols, all of these are listed as successive rows in the first
% column and they contain the same GO tags in the second column.

function [] = parse_GO_annots_file(go_consortium_annots_file, ...
                                   output_bp_annots_file, ...
                                   output_mf_annots_file, ...
                                   output_cc_annots_file)

annots_struct = goannotread(go_consortium_annots_file);

n_annots = size(annots_struct, 1);

fid_bp = fopen(output_bp_annots_file, 'w');
if (fid_bp == -1)
    disp('Cannot open BP annotations output file');
    return;
end

fid_mf = fopen(output_mf_annots_file, 'w');
if (fid_mf == -1)
    disp('Cannot open MF annotations output file');
    return;
end

fid_cc = fopen(output_cc_annots_file, 'w');
if (fid_cc == -1)
    disp('Cannot open CC annotations output file');
    return;
end

i = 1;
while (i <= n_annots)
    gene_ids = regexp(annots_struct(i).Synonym, '\|', 'split');

    n_gene_ids = length(gene_ids);
    std_id     = annots_struct(i).DB_Object_Symbol;
    id_exists  = 0;
    for j = 1:n_gene_ids
        if (strcmp(std_id, gene_ids{j}))
            id_exists = 1;
        end
    end
    if (id_exists == 0)
        gene_ids{n_gene_ids+1} = std_id;
        n_gene_ids             = n_gene_ids + 1;
    end

    ref_id   = gene_ids{1};
    cmp_id   = '';

    all_bp_go_ids_str = [];
    all_mf_go_ids_str = [];
    all_cc_go_ids_str = [];

    while ((i <= n_annots) & ...
           (isempty(cmp_id) | strcmp(ref_id, cmp_id)))
        
        if (strcmp(annots_struct(i).Aspect, 'P'))
            bp_go_id_str = sprintf('GO:%07d', annots_struct(i).GOid);           
            
            if(isempty(strfind(all_bp_go_ids_str, bp_go_id_str)))
                if (isempty(all_bp_go_ids_str))
                    all_bp_go_ids_str = strcat(all_bp_go_ids_str, bp_go_id_str);
                else
                    all_bp_go_ids_str = strcat(all_bp_go_ids_str, ';', bp_go_id_str);                
                end
            end
        end

        if (strcmp(annots_struct(i).Aspect, 'F'))
            mf_go_id_str = sprintf('GO:%07d', annots_struct(i).GOid);            
            
            if(isempty(strfind(all_mf_go_ids_str, mf_go_id_str)))
                if (isempty(all_mf_go_ids_str))
                    all_mf_go_ids_str = strcat(all_mf_go_ids_str, mf_go_id_str);
                else
                    all_mf_go_ids_str = strcat(all_mf_go_ids_str, ';', mf_go_id_str);                
                end
            end
        end

        if (strcmp(annots_struct(i).Aspect, 'C'))
            cc_go_id_str = sprintf('GO:%07d', annots_struct(i).GOid);            
            
            if(isempty(strfind(all_cc_go_ids_str, cc_go_id_str)))
                if (isempty(all_cc_go_ids_str))
                    all_cc_go_ids_str = strcat(all_cc_go_ids_str, cc_go_id_str);
                else
                    all_cc_go_ids_str = strcat(all_cc_go_ids_str, ';', cc_go_id_str);                
                end
            end
        end

        i = i + 1;
        
        if (i <= n_annots)
            next_gene_ids = regexp(annots_struct(i).Synonym, '\|', 'split');
            cmp_id        = next_gene_ids{1};
        end
    end

    % Debugging
    %i
    %gene_ids
    %all_bp_go_ids_str
    %all_mf_go_ids_str
    %all_cc_go_ids_str
    %pause;

    % If there are no GO tags from a particular ontology (BP, MF or CC) for a
    % gene, then assign it the root GO tag of the ontology.
    if (isempty(all_bp_go_ids_str))
        all_bp_go_ids_str = 'GO:0008150';
    end

    if (isempty(all_mf_go_ids_str))
        all_mf_go_ids_str = 'GO:0003674';
    end

    if (isempty(all_cc_go_ids_str))
        all_cc_go_ids_str = 'GO:0005575';
    end

    n_gene_ids = length(gene_ids);

    for j = 1:n_gene_ids
        count = fprintf(fid_bp, '%s\t%s\n', gene_ids{j}, all_bp_go_ids_str);
        if (count <= 0)
            disp('Cannot write to BP annotations output file');
            return;
        end

        count = fprintf(fid_mf, '%s\t%s\n', gene_ids{j}, all_mf_go_ids_str);
        if (count <= 0)
            disp('Cannot write to MF annotations output file');
            return;
        end

        count = fprintf(fid_cc, '%s\t%s\n', gene_ids{j}, all_cc_go_ids_str);
        if (count <= 0)
            disp('Cannot write to CC annotations output file');
            return;
        end
    end
end
