#!/usr/local/bin/perl

# Create a matrix for input to programs using GO tags. 
#
# Each row of this matrix corresponds to a gene with the first column specifying
# the gene identity and the remaining columns specifying the associated ontology
# term(s') identity(ies). Note that the identities are all integer numbers.
# Further, the ontology term identities are serial integers but the gene
# identities are necessarily not. Since different genes have
# different number of GO annotations, the number of GO term columns for each
# gene is equal to the maximum number of annotations for the genes in the set.
# The columns for genes that do not have the maximum number of annotations are
# appended with -1's toward the end.
# 
# The argument 'all_go_tags_file' specifies the name of a file containing gene
# IDs (alphanumeric) and the corresponding GO tags (GO:xxxxxxx) as the first and
# second columns respectively. It is usually derived from the GO tag files
# available for model organisms. So the set of genes contained in this file is
# usually a superset of the genes being studied in a microarray experiment.
#
# 'gene_ids_file' is the name of the file containing the IDs (alphanumeric) of
# the genes for which the matrix is to be built. This ID list is usually a subset
# of all the IDs in the first column of the 'all_go_tags_file'. It should be a
# two column file with the first column specifying any unique integer IDs for the
# genes and the second column their alphanumeric IDs. If there are no unique
# integer IDs, then serial numbers are OK.
#
# 'serial_go_ids_file' specifies the name of a file containing serial integer
# IDs for the GO tags of all the genes under study. The first column of the
# file lists the GO tags (GO:xxxxxxx) and the second column the corresponding
# serial IDs. If this file exists, then the integer IDs are used to create the
# GO tag columns of the output file. Otherwise, it is created by the script
# anyway and can be used for future reference.
#
# The 'root_go_id' argument specifies the GO ID (GO:xxxxxxx) of the root of the
# particular GO graph from which the terms are being used. It is one of the
# following:
# Biological Process: GO:0008150 
# Molecular function: GO:0003674
# Cellular Component: GO:0005575
#
# 'output_file' specifies the name of the output file into which the formatted
# matrix of serial integre GO IDs for each gene is written out. 
 
if ($#ARGV < 4)
{
    die "Usage: $0 all_go_tags_file gene_ids_file serial_go_ids_file root_go_id output_file\n";
}

($all_go_tags_file, $gene_ids_file, $serial_go_ids_file, $root_go_id, $output_file) = @ARGV;

# Using all the genes that have any GO tag:
# Build a "%gene_id_hash" table with gene IDs as the keys and the GO tags as
# the values in the table.
open(ALL_GO_TAGS_FILE, "<$all_go_tags_file") || die "Cannot open file $all_go_tags_file: $!\n";

$go_tag_id = 1;

while (<ALL_GO_TAGS_FILE>)
{
    @line = split();

    $gene_id_hash{lc($line[0])} = $line[1];

    #@fld_2 = split(/;/, $line[1]);
    #
    #foreach $j (0..$#fld_2)
    #{
    #    @exist_go_tags  = keys %go_tags_hash;
    #    $already_exists = 0;
    #    
    #    foreach $i (0..$#exist_go_tags)
    #    {
    #        if (lc($fld_2[$j]) eq lc($exist_go_tags[$i]))
    #        {
    #            $already_exists = 1;
    #        }
    #    }
    #
    #    if ($already_exists == 0)
    #    {
    #        $go_tags_hash{lc($fld_2[$j])} = $go_tag_id;
    #        $go_tag_id++;
    #    }
    #}
}
close ALL_GO_TAGS_FILE;

# Debugging.
#foreach $key (keys %gene_id_hash)
#{
#    print $key, "\t", $gene_id_hash{lc($key)}, "\n";
#}

# Build a "%req_go_tags_hash" table for all the genes under study. This is
# different from the "%go_tags_hash" table in that it uses only a subset of the
# genes, i.e., those specified by the 'gene_ids_file'.
# The "%gene_id_hash" table is used to access the GO tags for these genes.

if ( -e $serial_go_ids_file )
{
    $serial_go_ids_file_exists = 1;
    open (SERIAL_GO_IDS_FILE, "<$serial_go_ids_file") || die "Cannot open file $serial_go_ids_file for reading: $!\n";

    # Build the "%req_go_tags_hash" table from the file.
    while (<SERIAL_GO_IDS_FILE>)
    {
        @line = split();

        $req_go_tags_hash{lc($line[0])} = $line[1];
    }
}
else
{
    $serial_go_ids_file_exists = 0;
    open (SERIAL_GO_IDS_FILE, ">$serial_go_ids_file") || die "Cannot open file $serial_go_ids_file for writing: $!\n";
}

# Build the "%req_go_tags_hash" table from the gene IDs file if the
# "$serial_go_ids_file" does not exist.
if ($serial_go_ids_file_exists == 0)
{
    $go_tag_id = 1;

    open (GENE_IDS_FILE, "<$gene_ids_file") || die "Cannot open file $gene_ids_file for reading: $!\n";

    while (<GENE_IDS_FILE>)
    {
        @line = split();

        # Debugging.
        #print $line[0], "\t", $cl, "\t"; 
        
        if ($#line < 1)
        {
            $go_tags = $root_go_id;
        }
        else
        {
            $go_tags = $gene_id_hash{lc($line[1])};

            # For non-existent ids in the gene_id_hash table, we put
            # the root_go_id
            if ($go_tags eq '')
            {
                $go_tags = $root_go_id;
            }
        }
        
        # Debugging.
        #print $go_tags, "\n";

        @go_tags = split(/;/, $go_tags);
        
        foreach $j (0..$#go_tags)
        {
        
            @exist_go_tags  = keys %req_go_tags_hash;
            $already_exists = 0;
            
            foreach $i (0..$#exist_go_tags)
            {
                if (lc($go_tags[$j]) eq lc($exist_go_tags[$i]))
                {
                    $already_exists = 1;
                }
            }

            if ($already_exists == 0)
            {
                $req_go_tags_hash{lc($go_tags[$j])} = $go_tag_id;
                $go_tag_id++;
            }
                        
            #print $req_go_tags_hash{$go_tags[$j]}, "\t";
        }
        
        #print "\n";

        if ($#go_tags > $max_num_go_tags)
        {
            $max_num_go_tags = $#go_tags;
        }
    }

    close GENE_IDS_FILE;

    foreach $key (sort { $req_go_tags_hash{lc($a)} <=> $req_go_tags_hash{lc($b)} } keys %req_go_tags_hash)
    {
        print SERIAL_GO_IDS_FILE uc($key), "\t", $req_go_tags_hash{lc($key)}, "\n";
    }
}
else
{
    open (GENE_IDS_FILE, "<$gene_ids_file") || die "Cannot open file $gene_ids_file for reading: $!\n";

    while (<GENE_IDS_FILE>)
    {
        @line = split();

        if ($#line < 1)
        {
            $go_tags = $root_go_id;
        }
        else
        {
            $go_tags = $gene_id_hash{lc($line[1])};
            
            # For non-existent ids in the gene_id_hash table, we put
            # the root_go_id
            if ($go_tags eq '')
            {
                $go_tags = $root_go_id;
            }
        }
        
        @go_tags = split(/;/, $go_tags);
        
        if ($#go_tags > $max_num_go_tags)
        {
            $max_num_go_tags = $#go_tags;
        }
    }

    close GENE_IDS_FILE;
}

# Debugging.
print "Maximum # GO tags: ", $max_num_go_tags, "\n";
#
#foreach $key (keys %req_go_tags_hash)
#{
#    print $key, "\t", $req_go_tags_hash{lc($key)}, "\n";
#}

# Print out the formatted data.
open (OUT_FILE, ">$output_file") || die "Cannot open file $output_file for writing: $!\n";
open (GENE_IDS_FILE, "<$gene_ids_file") || die "Cannot open file $gene_ids_file for reading: $!\n";

while (<GENE_IDS_FILE>)
{
    @line = split();

    #print $line[0], "\t", $cl, "\t"; 
    print OUT_FILE $line[0], "\t"; 

    if ($#line < 1)
    {
        $go_tags = $root_go_id;
    }
    else
    {
        $go_tags = $gene_id_hash{lc($line[1])};

        # For non-existent ids in the gene_id_hash table, we put
        # the root_go_id
        if ($go_tags eq '')
        {
            $go_tags = $root_go_id;
        }
    }

    @go_tags = split(/;/, $go_tags);

    foreach $j (0..$max_num_go_tags)
    {            
        if ($j <= $#go_tags)
        {
            print OUT_FILE $req_go_tags_hash{lc($go_tags[$j])}, "\t";
        }
        else
        {                
            print OUT_FILE "-1", "\t";
        }
    }        
    print OUT_FILE "\n";
}

close GENE_IDS_FILE;
close OUT_FILE;
