% Predict label for test sample(s) based on maximum likelihood of the generative
% model for clustering gene expressions and GO tags together. This routine
% computes only probabilities based on likelihoods and returns them. The actual
% prediction based on maximum likelihood can be done using the returned values.
%
% The generative model implies a joint distribution over the space of gene
% expression levels and GO tags in a parametric form. The model parameters are
% estimated using a learning algorithm that optimizes a certain objective
% function, eg. likelihood or posterior. This joint distribution is used to
% predict the time (sample) or stage label for a given test sample. Time
% (sample) label prediction is done by computing marginal likelihood of the gene
% expression measurements of the test sample under the learnt joint distribution. Stage
% label prediction is done by either using the marginal time (sample) label
% probabilities with the knowledge of time-to-stage mapping or directly using a
% Bayesian maximum likelihood framework. The latter is possible only with models
% which have stage information built in as part of the learning procedure.
%
% The input parameter 'train_samples_file' is a file of gene expression vectors
% used for model training. Each row corresponds to a gene and each column
% corresponds to a sample or time-point.
%
% 'go_tags_file' is the name of a file containing the matrix of integer IDs of
% GO tags for the genes. Each row corresponds to a gene and each column to one
% of its GO tags. Since different genes have different numbers of annotations,
% the number of columns in the matrix is equal to the maximum number of
% annotations for any gene. The columns for each gene are first filled with
% positive integers corresponding to its GO tag IDs followed by -1's if the gene
% does not have the maximum number of annotations.The first column is an integer
% ID for the gene itself.
%
% 'model_mat_file' is the name of a .mat file into which model parameters are
% written. These include the means and variances of the Gaussian mixture and the
% estimated joint distribution P(o, c) over ontology terms and clusters.
%
% The option 'do_marginalize' specifies whether marginal likelihoods are to be
% used for computing the sample or time label probabilities. Otherwise, joint
% likelihoods are used by using training data for the other sample or time
% labels. In other words, the test sample expression values are substituted only
% at the sample or time point for which the probability is being computed and
% training data is used as is for the others and the joint likelihood is
% computed. The marginalization is performed only when this option is
% non-zero.
%
% 'do_weight_GO_tags' specifies whether the weights P(g|o) have to be considered
% or not. When it is set to 0, P(g|o) is 1.0 everywhere, otherwise it is a
% uniform distribution over all genes 'g' that are annotated with the term 'o'.
% Please refer to the documentation of the corresponding model training routine
% for more details about P(g|o).
%
% The option 'ignore_go_tags', when set to a non-zero value, causes the GO tags
% associated with genes to be ignored when computing the marginal or joint
% likelihoods. Otherwise they are used in computing the likelihoods and thereby
% the sample, time or state probabilities.
%
% The option 'do_gene_identifiability' forces gene identifiability during 
% stage prediction by using gene specific cluster posteriors instead of the
% overall cluster posteriors in the likelihood computations.
%
% The output parameter 'pred_probs' is a matrix of probabilities of time
% (sample) labels for the test samples. Each row corresponds to a test sample
% of the data in the 'test_samples_file' and each column corresponds to one of
% the possible time (sample) labels. So it is a matrix of dimension
% num-test-samples-BY-num-train-samples. 

function [pred_probs] = predict_max_lhood_label(train_samples_file, ...
                                                test_samples_file, ...
                                                go_tags_file, ...
                                                model_mat_file, ...
                                                do_marginalize, ...
                                                do_weight_GO_tags, ...
                                                ignore_GO_tags, ...
                                                n_closest_genes, ...
                                                do_weighted_expr_avg, ...
                                                do_direct_inference, ...
                                                do_gene_identifiability)

load(model_mat_file);

all_train_gene_exprs = load(train_samples_file);
all_test_gene_exprs  = load(test_samples_file);
all_go_tags_data     = load(go_tags_file);

% Maximum number of genes is limited by the available data.
n_closest_genes = min(n_closest_genes, size(all_train_gene_exprs, 1));

if (n_closest_genes)
    closest_gene_idxs = unique(closest_genes_g_c(1:n_closest_genes, :))
    
    train_gene_exprs = all_train_gene_exprs(closest_gene_idxs, :);    
    test_gene_exprs  = all_test_gene_exprs(closest_gene_idxs, :);    
    go_tags_data     = all_go_tags_data(closest_gene_idxs, :);
else
    train_gene_exprs = all_train_gene_exprs;    
    test_gene_exprs  = all_test_gene_exprs;    
    go_tags_data     = all_go_tags_data;
end

n_genes         = size(train_gene_exprs, 1);
n_train_samples = size(train_gene_exprs, 2);
n_test_samples  = size(test_gene_exprs, 2);

n_fields  = size(go_tags_data, 2);
go_tags   = go_tags_data(:, 2:n_fields);
n_go_tags = max(go_tags(:))

n_clusters = size(mean_c, 1);

% A variable with name p_c_bar_O_g gets initialized to zero in order to be used
% later. So copy it to form the weights before it gets modified.
log_weights  = zeros(n_genes, n_clusters); 
norm_weights = zeros(n_genes, n_clusters); 

if (ignore_GO_tags)
    sum_weights  = sum(p_c_bar_O_g, 1)
end

for c = 1:n_clusters
    if (ignore_GO_tags)
        if (sum_weights(c) > 0)
            norm_weights(:, c) = p_c_bar_O_g(:, c) ./ sum_weights(c);
        end
    else
        for g = 1:n_genes
            log_weights(g, c) = safe_log(p_c_bar_O_g(g, c)) + log_p_O_bar_c(g, c);
        end
        [norm_weights(:, c), sum_log] = scale_by_sum_log_vec(log_weights(:, c));
    end
end

weighted_expr_avg = zeros(n_clusters, 1);
log_p_G_bar_c     = zeros(n_clusters, 1);
log_p_G_c         = zeros(n_clusters, 1);
p_c_bar_G         = zeros(n_clusters, 1);

% Useful data structures.
% Make a copy of p_c_bar_O_g first.
p_c_bar_O_g_model = p_c_bar_O_g;
p_c_bar_O_g       = zeros(n_genes, n_clusters);
p_g_bar_o         = zeros(n_genes, n_go_tags);
p_g_bar_c         = zeros(n_genes, n_clusters);
log_prod_stdev_c  = zeros(n_clusters, 1); 
log_p_g_bar_c     = zeros(n_genes, n_clusters);
log_p_g_o_c       = zeros(n_clusters, n_go_tags, n_genes);
log_p_g_O_c       = zeros(n_genes, n_clusters);
log_p_tilda_g_O_c = zeros(n_genes, n_clusters);
log_p_o_c         = zeros(n_clusters, n_go_tags);
p_c               = zeros(n_clusters, 1);
log_p_c           = zeros(n_clusters, 1);
M_g               = zeros(n_genes, 1);

% Inference by averaging log(s|gene) over genes.
log_p_g_O_c_bar_g       = zeros(n_genes, n_clusters, n_train_samples);
p_c_bar_O_g_bar_g       = zeros(n_genes, n_clusters, n_train_samples);
log_p_g_O_s_bar_g       = zeros(n_genes, n_train_samples);
log_p_tilda_g_O_s_bar_g = zeros(n_genes, n_train_samples);
p_s_bar_O_g             = zeros(n_genes, n_train_samples);

mixed_gene_exprs     = zeros(n_genes, n_train_samples);
log_likelihood       = zeros(n_test_samples, n_train_samples);
log_tilda_likelihood = zeros(n_test_samples, n_train_samples);
pred_probs           = zeros(n_test_samples, n_train_samples);
M_t                  = zeros(n_test_samples, 1);

% For direct inference.
%p_c_bar_O_g_train       = zeros(n_genes, n_clusters);
p_g_bar_c_train         = zeros(n_genes, n_clusters);
log_p_g_bar_c_train     = zeros(n_genes, n_clusters);
log_p_g_O_c_train       = zeros(n_genes, n_clusters);
%log_p_tilda_g_O_c_train = zeros(n_genes, n_clusters);
log_p_g_O_bar_c         = zeros(n_genes, n_clusters);
log_p_c_bar_s           = zeros(n_clusters, 1); 
p_c_bar_s               = zeros(n_clusters, 1); 
log_p_g_O_c_bar_s       = zeros(n_genes, n_clusters);
log_p_tilda_g_O_c_bar_s = zeros(n_genes, n_clusters);
p_g_O_bar_s             = zeros(n_genes, n_clusters);
p_c_bar_O_G_train       = zeros(n_clusters, 1);
log_p_G_O_c_train       = zeros(n_clusters, 1);
log_p_tilda_G_O_c_train = zeros(n_clusters, 1);
M_G                     = 0;

% p_g_bar_o is uniform over all the genes annotated with the term 'o'.
go_tag_probs = zeros(n_go_tags, 1);
for o = 1:n_go_tags
    o_go_genes = length(find(go_tags(:) == o));
    if (o_go_genes > 0)
        go_tag_probs(o) = 1.0/o_go_genes;
    end
end

n_go_tags_g = zeros(n_genes, 1);
for g = 1:n_genes
    g_go_tags      = go_tags(g, :);
    n_go_tags_g(g) = length(find(g_go_tags > 0));

    for o = 1:n_go_tags
        o_go_tag_idxs = find(g_go_tags == o);

        if( ~isempty(o_go_tag_idxs) )
            if (do_weight_GO_tags)
                p_g_bar_o(g, o) = go_tag_probs(o);
            else
                p_g_bar_o(g, o) = 1.0;
            end
        end
    end
end

% Form p_c, log_p_c and log_p_o_c from p_o_c.
p_c = sum(p_o_c, 2);
for c = 1:n_clusters
    log_p_c(c) = safe_log(p_c(c));

    for o = 1:n_go_tags
        log_p_o_c(c, o) = safe_log(p_o_c(c, o));
    end
end

for test_s = 1:n_test_samples    
    for train_s = 1:n_train_samples

        mixed_gene_exprs             = train_gene_exprs;
        mixed_gene_exprs(:, train_s) = test_gene_exprs(:, test_s);
    
        stdev_c = sqrt(var_c);
        
        if (do_weighted_expr_avg)

            for c = 1:n_clusters
                
                weighted_expr_avg(c) = dot(norm_weights(:, c), test_gene_exprs(:, test_s));

                log_p_G_bar_c(c) = -0.5*(weighted_expr_avg(c) - mean_c(c, train_s))*(weighted_expr_avg(c) - mean_c(c, train_s))/var_c(c, train_s);
                log_p_G_bar_c(c) = log_p_G_bar_c(c) - safe_log(stdev_c(c, train_s));

                log_p_G_c(c) = log_p_G_bar_c(c) + log_p_c(c);

            end

            [p_c_bar_G, log_likelihood(test_s, train_s)] = scale_by_sum_log_vec(log_p_G_c); 

        else

            %%%%%%%% Compute log likelihood %%%%%%%%
            p_c_bar_O_g       = zeros(n_genes, n_clusters);
            %p_c_bar_O_g_train = zeros(n_genes, n_clusters);

            % Form log_p_g_o_c
            for c = 1:n_clusters
                
                log_prod_stdev_c(c) = 0;

                if (do_marginalize)
                    log_prod_stdev_c(c) = log_prod_stdev_c(c) + safe_log(stdev_c(c, train_s));
                else
                    for s = 1:n_train_samples
                        log_prod_stdev_c(c) = log_prod_stdev_c(c) + safe_log(stdev_c(c, s));
                    end
                end
                % Debugging.
                %log_prod_stdev_c;

                for g = 1:n_genes                
                    if (do_direct_inference)                  
                        log_p_g_bar_c_train(g, c) = -0.5*sum(((train_gene_exprs(g, train_s) - mean_c(c, train_s)).*(train_gene_exprs(g, train_s) - mean_c(c, train_s)))./var_c(c, train_s));
                        log_p_g_bar_c_train(g, c) = log_p_g_bar_c_train(g, c) - log_prod_stdev_c(c);
                    end

                    if (do_marginalize || do_direct_inference)
                        log_p_g_bar_c(g, c) = -0.5*sum(((mixed_gene_exprs(g, train_s) - mean_c(c, train_s)).*(mixed_gene_exprs(g, train_s) - mean_c(c, train_s)))./var_c(c, train_s));
                    else
                        log_p_g_bar_c(g, c) = -0.5*sum(((mixed_gene_exprs(g, :) - mean_c(c, :)).*(mixed_gene_exprs(g, :) - mean_c(c, :)))./var_c(c, :));
                    end
                    log_p_g_bar_c(g, c) = log_p_g_bar_c(g, c) - log_prod_stdev_c(c);

                    g_go_tags = go_tags(g, :);
                    
                    % Too slow!
                    %for o = 1:n_go_tags
                    %    % Do it only for valid ontologies for a gene.
                    %    o_go_tag_idxs = find(g_go_tags == o);
                    %    if( ~isempty(o_go_tag_idxs) )
                    %        if (ignore_GO_tags)
                    %            log_p_g_o_c(c, o, g) = log_p_g_bar_c(g, c) + log_p_c(c); 
                    %        else
                    %            log_p_g_o_c(c, o, g) = safe_log(p_g_bar_o(g, o)) + log_p_g_bar_c(g, c); 
                    %            log_p_g_o_c(c, o, g) = log_p_g_o_c(c, o, g) + log_p_o_c(c, o); 
                    %        end
                    %    end
                    %end

                    % Relatively faster.
                    for o_i = 1:n_go_tags_g(g)
                        o = g_go_tags(o_i);
                        if (ignore_GO_tags)
                            if (do_direct_inference)
                                log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) + log_p_g_bar_c_train(g, c);
                                log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) + log_p_c(c);
                            end

                            log_p_g_o_c(c, o, g) = log_p_g_bar_c(g, c) + log_p_c(c); 
                        else
                            if (do_direct_inference)
                                log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) + safe_log(p_g_bar_o(g, o));
                                log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) + log_p_g_bar_c_train(g, c);
                                log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) + log_p_o_c(c, o);
                            end

                            log_p_g_o_c(c, o, g) = safe_log(p_g_bar_o(g, o)) + log_p_g_bar_c(g, c); 
                            log_p_g_o_c(c, o, g) = log_p_g_o_c(c, o, g) + log_p_o_c(c, o); 
                        end
                    end

                    if (do_direct_inference)
                        log_p_g_O_c_train(g, c) = log_p_g_O_c_train(g, c) ./ n_go_tags_g(g);                                 
                    end

                    % Doesn't seem quite right:
                    %log_p_g_o_c(:, :, g) = log_p_g_o_c(:, :, g) ./ n_go_tags_g(g);
                    % Should be more like:
                    log_p_g_o_c(c, :, g) = log_p_g_o_c(c, :, g) ./ n_go_tags_g(g);
                end

            end

            % Form log_p_g_O_c and log_p_tilda_g_O_c.
            for g = 1:n_genes
                log_p_g_O_c(g, :) = (sum(log_p_g_o_c(:, :, g), 2))';

                % Compute log_p_g_O_bar_c.
                for c = 1:n_clusters
                    log_p_g_O_bar_c(g, c) = log_p_g_O_c(g, c) - log_p_c(c);    
                end

            end
            
            % Normalize and compute log likelihood.
            if (do_direct_inference)

                % Average over genes. 
                %sum_over_c = zeros(n_genes, 1);
		%for g = 1:n_genes
		%    M_g(g)                        = max(log_p_g_O_c_train(g, :));
		%    log_p_tilda_g_O_c_train(g, :) = log_p_g_O_c_train(g, :) - M_g(g);
		%    p_c_bar_O_g_train(g, :)       = exp(log_p_tilda_g_O_c_train(g, :)); 
		%    sum_over_c(g)                 = sum(p_c_bar_O_g_train(g, :));
		%    p_c_bar_O_g_train(g, :)       = p_c_bar_O_g_train(g, :) ./ sum_over_c(g);
		%end                
                %p_c_bar_s = (sum(p_c_bar_O_g_train(:, :), 1))';
                %p_c_bar_s = p_c_bar_s ./ n_genes;
                
                % Product over genes.
                %log_p_G_O_c_train       = (sum(log_p_g_O_c_train(:, :), 1))';
                %M_G                     = max(log_p_G_O_c_train);
                %log_p_tilda_G_O_c_train = log_p_G_O_c_train - M_G;   
                %p_c_bar_O_G_train       = exp(log_p_tilda_G_O_c_train);
                %sum_over_c_G            = sum(p_c_bar_O_G_train);
		%p_c_bar_O_G_train       = p_c_bar_O_G_train ./ sum_over_c_G;
                %p_c_bar_s               = p_c_bar_O_G_train;

                % Compute p_g_O_bar_s and log likelihood (p_c_bar_s comes from either average or product over genes).
                %sum_over_c = zeros(n_genes, 1);
                %for g = 1:n_genes
                %    log_p_g_O_c_bar_s(g, :) = log_p_g_O_bar_c(g, :) + (safe_log(p_c_bar_s))';
                %
                %    M_g(g)                        = max(log_p_g_O_c_bar_s(g, :));
                %    log_p_tilda_g_O_c_bar_s(g, :) = log_p_g_O_c_bar_s(g, :) - M_g(g);
                %    p_g_O_c_bar_s(g, :)           = exp(log_p_tilda_g_O_c_bar_s(g, :));
                %    sum_over_c(g)                 = sum(p_g_O_c_bar_s(g, :));
                %
                %    log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + log(sum_over_c(g)) + M_g(g);
                %end

                % Inference by averaging log(s|gene) over genes.
                log_p_g_O_c_bar_g(g, :, train_s) = log_p_g_O_bar_c(g, :) + safe_log(p_c_bar_O_g_model(g, :));
                [p_c_bar_O_g_bar_g, log_p_g_O_s_bar_g(g, train_s)] = scale_by_sum_log_vec(log_p_g_O_c_bar_g(g, :, train_s));

            else

                % Use gene specific cluster responsibilities.
                if (do_gene_identifiability)
                    log_p_g_O_c(g, :) = log_p_g_O_bar_c(g, :) + safe_log(p_c_bar_O_g_model(g, :));
                end
 
                sum_over_c = zeros(n_genes, 1);
                for g = 1:n_genes
                    M_g(g)                  = max(log_p_g_O_c(g, :));
                    log_p_tilda_g_O_c(g, :) = log_p_g_O_c(g, :) - M_g(g);
                    p_c_bar_O_g(g, :)       = exp(log_p_tilda_g_O_c(g, :)); 
                    sum_over_c(g)           = sum(p_c_bar_O_g(g, :));
                    p_c_bar_O_g(g, :)       = p_c_bar_O_g(g, :) ./ sum_over_c(g);

                    log_likelihood(test_s, train_s) = log_likelihood(test_s, train_s) + log(sum_over_c(g)) + M_g(g);
                end
            end

            % Debugging.
            %M_g;
            %sum_over_c;
            %log_likelihood;

        end

    end

    if (do_direct_inference)
        % Inference by averaging log(s|gene) over genes.
        sum_over_train_s = zeros(n_genes, 1);
        for g = 1:n_genes
            M_g(g)                        = max(log_p_g_O_s_bar_g(g, :));
            log_p_tilda_g_O_s_bar_g(g, :) = log_p_g_O_s_bar_g(g, :) - M_g(g);
            p_s_bar_O_g(g, :)             = exp(log_p_tilda_g_O_s_bar_g(g, :)); 
            sum_over_train_s(g)           = sum(p_s_bar_O_g(g, :));
            p_s_bar_O_g(g, :)             = p_s_bar_O_g(g, :) ./ sum_over_train_s(g);
        end
        for train_s = 1:n_train_samples
            log_likelihood(test_s, train_s) = safe_log(mean(p_s_bar_O_g(:, train_s)));
        end
    end

end
% Debugging.
%log_likelihood

% Convert log likelihoods to prediction probabilities.
for test_s = 1:n_test_samples
    M_t(test_s)                     = max(log_likelihood(test_s, :));
    log_tilda_likelihood(test_s, :) = log_likelihood(test_s, :) - M_t(test_s);
end

sum_over_s = zeros(n_test_samples, 1);
for test_s = 1:n_test_samples
    pred_probs(test_s, :) = exp(log_tilda_likelihood(test_s, :));
    sum_over_s(test_s)    = sum(pred_probs(test_s, :));
    pred_probs(test_s, :) = pred_probs(test_s, :) ./ sum_over_s(test_s);
end
