% Predict label for test sample(s) based on maximum likelihood of the generative
% model for state or pool-specific clustering of gene expressions and GO tags
% together. This routine computes only probabilities based on likelihoods and
% returns them. The actual prediction based on maximum likelihood can be done
% using the returned values.
%
% The generative model implies a joint distribution over the space of gene
% expression levels and GO tags in a parametric form. The model parameters are
% estimated using a learning algorithm that optimizes a certain objective
% function, eg. likelihood or posterior. This joint distribution is used to
% predict the time (sample) or stage label for a given test sample. Time
% (sample) label prediction is done by computing marginal likelihood of the gene
% expression measurements of the test sample under the learnt joint
% distribution. Stage label prediction is done by either using the marginal time
% (sample) label probabilities with the knowledge of time-to-stage mapping or
% directly using a Bayesian maximum likelihood framework. The latter is possible
% only with models, which have stage information built in as part of the learning
% procedure.
%
% The input parameter 'train_samples_file' is a file of gene expression vectors
% used for model training. Each row corresponds to a gene and each column
% corresponds to a sample or time-point.
%
% 'go_tags_file' is the name of a file containing the matrix of integer IDs of
% GO tags for the genes. Each row corresponds to a gene and each column to one
% of its GO tags. Since different genes have different numbers of annotations,
% the number of columns in the matrix is equal to the maximum number of
% annotations for any gene. The columns for each gene are first filled with
% positive integers corresponding to its GO tag IDs followed by -1's if the gene
% does not have the maximum number of annotations.The first column is an integer
% ID for the gene itself.
%
% 'model_mat_file' is the name of a .mat file into which model parameters are
% written. These include the means and variances of the Gaussian mixtures of
% each sample or time-point and the estimated joint distributions P(o, c_s) over
% ontology terms and clusters.
%
% The option 'do_marginalize' specifies whether marginal likelihoods are to be
% used for computing the sample or time label probabilities. Otherwise, joint
% likelihoods are used by using training data for the other sample or time
% labels. In other words, the test sample expression values are substituted only
% at the sample or time point for which the probability is being computed and
% training data is used as is for the others and the joint likelihood is
% computed. The marginalization is performed only when this option is
% non-zero.
%
% 'do_weight_GO_tags' specifies whether the weights P(g|o) have to be considered
% or not. When it is set to 0, P(g|o) is 1.0 everywhere, otherwise it is a
% uniform distribution over all genes 'g' that are annotated with the term 'o'.
% Please refer to the documentation of the corresponding model training routine
% for more details about P(g|o).
%
% The option 'ignore_go_tags', when set to a non-zero value, causes the GO tags
% associated with genes to be ignored when computing the marginal or joint
% likelihoods. Otherwise they are used in computing the likelihoods and thereby
% the sample, time or state probabilities.
%
% The option 'do_gene_identifiability' forces gene identifiability during 
% stage prediction by using gene specific cluster posteriors instead of the
% overall cluster posteriors for each time point in the likelihood computations.
%
% The output parameter 'time_pred_probs' is a matrix of probabilities of time
% (sample) labels for the test samples. Each row corresponds to a test sample
% of the data in the 'test_samples_file' and each column corresponds to one of
% the possible time (sample) labels. So it is a matrix of dimension
% num-test-samples-BY-num-train-samples.
%
% The output parameter 'state_pred_probs' is a matrix of probabilities of state
% labels for the test samples. Each row corresponds to a test sample
% of the data in the 'test_samples_file' and each column corresponds to one of
% the possible state labels. So it is a matrix of dimension
% num-test-samples-BY-num-states. These are computed using a
% Bayesian framework. 

function [time_pred_probs, state_pred_probs] = predict_max_lhood_label_state_specific_dependent_tied_params(train_samples_file, ...
                                                                                                            test_samples_file, ...
                                                                                                            go_tags_file, ...
                                                                                                            model_mat_file, ...
                                                                                                            do_marginalize, ...
                                                                                                            do_weight_GO_tags, ...
                                                                                                            ignore_GO_tags, ...
                                                                                                            do_count_GO_tags_once, ...
                                                                                                            do_gene_identifiability, ...
                                                                                                            do_arithmetic_avg)

% Direct state prediction (p(s|D) propto p(D|s)).
do_skip_time_points = 1;

load(model_mat_file);

all_train_gene_exprs = load(train_samples_file);
all_test_gene_exprs  = load(test_samples_file);
all_go_tags_data     = load(go_tags_file);

% Maximum number of genes is limited by the available data.
%n_closest_genes = min(n_closest_genes, size(all_train_gene_exprs, 1));
%
%if (n_closest_genes)
%    closest_gene_idxs = unique(closest_genes_g_c(:, 1, 1:n_closest_genes, :))
%
%    train_gene_exprs = all_train_gene_exprs(closest_gene_idxs, :);    
%    test_gene_exprs  = all_test_gene_exprs(closest_gene_idxs, :);    
%    go_tags_data     = all_go_tags_data(closest_gene_idxs, :);
%else
    train_gene_exprs = all_train_gene_exprs;    
    test_gene_exprs  = all_test_gene_exprs;    
    go_tags_data     = all_go_tags_data;
%end

n_train_samples = size(train_gene_exprs, 2);
n_test_samples  = size(test_gene_exprs, 2);

% Useful variables.
total_num_feature_vectors = size(train_gene_exprs, 1);
num_exprs_per_gene        = 1;
n_genes                   = total_num_feature_vectors/num_exprs_per_gene;

n_fields          = size(go_tags_data, 2);
num_tags_per_gene = n_fields - 1;

% Reshape the feature data into a 3-dimensional matrix such that:
% 1. each plane corresponds to an gene.
% 2. in each plane, rows correspond to regions and columns correspond to
% feature dimensions.
train_feature = zeros(num_exprs_per_gene, n_train_samples, n_genes);
test_feature  = zeros(num_exprs_per_gene, n_test_samples, n_genes);
mixed_feature = zeros(num_exprs_per_gene, n_train_samples, n_genes);
for i = 1:n_genes
    train_feature(:, :, i) = train_gene_exprs((num_exprs_per_gene*(i - 1) + 1):(num_exprs_per_gene*i), :);
    test_feature(:, :, i)  = test_gene_exprs((num_exprs_per_gene*(i - 1) + 1):(num_exprs_per_gene*i), :);
end

word        = zeros(num_tags_per_gene, 1, n_genes);
n_go_tags_g = zeros(n_genes, 1);
for i = 1:n_genes
    word(:, 1, i)  = (go_tags_data(i, 2:n_fields))';
    n_go_tags_g(i) = length(find(word(:, 1, i) > 0));
end

n_states  = length(n_clusters);
n_go_tags = max(word(:))
M         = max(n_clusters);

%n_clusters = size(mean_c, 1);
max_n_clusters = max(n_clusters);

log_likelihood       = zeros(n_test_samples, n_train_samples);
log_tilda_likelihood = zeros(n_test_samples, n_train_samples);
time_pred_probs      = zeros(n_test_samples, n_train_samples);
state_pred_probs     = zeros(n_test_samples, n_states);
M_t                  = zeros(n_test_samples, 1);

% Useful data structures.
P_c_given_x_i     = zeros(num_exprs_per_gene, M, n_genes, n_train_samples);
P_c_given_w_i     = zeros(num_tags_per_gene, M, n_genes, n_train_samples);
sum_P_c_given_x_i = zeros(M, n_train_samples);
P_c_given_X_i     = zeros(M, 1, n_genes, n_train_samples);

log_prod_stdev_c = zeros(M, n_states); 
log_P_x_c        = zeros(M, num_exprs_per_gene, n_genes, n_train_samples);
log_P_tilda_x_c  = zeros(M, num_exprs_per_gene, n_genes, n_train_samples);
log_P_c          = zeros(M, n_states);
M_x              = zeros(num_exprs_per_gene, 1, n_genes, n_train_samples);         

log_P_w_given_X = zeros(num_tags_per_gene, 1, n_genes, n_train_samples); 
log_P_tilda_x   = zeros(num_exprs_per_gene, 1, n_genes, n_train_samples);

% State likelihood computation.
log_P_s             = zeros(n_states, 1);
P_c_s_given_x_i     = zeros(num_exprs_per_gene, M, n_genes);
P_c_s_given_w_i     = zeros(num_tags_per_gene, M, n_genes);
sum_P_c_s_given_x_i = zeros(M, 1);
P_c_s_given_X_i     = zeros(M, 1, n_genes);

log_P_x_c_s        = zeros(M, num_exprs_per_gene, n_genes);
log_P_tilda_x_c_s  = zeros(M, num_exprs_per_gene, n_genes);
M_x_s              = zeros(num_exprs_per_gene, 1, n_genes);         

log_P_w_given_X_s = zeros(num_tags_per_gene, 1, n_genes); 
log_P_tilda_x_s   = zeros(num_exprs_per_gene, 1, n_genes);

% Gene identifiability.
log_P_c_given_X_g = zeros(M, n_genes, n_states);

% Form stdev_c, log_product_stdev_c, and log_P_c.
stdev_c = sqrt(v_c);
for s = 1:n_states
    for c = 1:M
        log_product_stdev_c(c, s) = safe_log(prod(stdev_c(c, s)));
    end
    log_P_c(:, s) = safe_log(P_c(:, s));

    if (do_gene_identifiability)
        for i = 1:n_genes
            log_P_c_given_X_g(:, i, s) = safe_log(P_c_given_X_g(:, i, s)); 
        end
    end
end

for test_t = 1:n_test_samples  
    
    for test_s = 1:n_states

        if ( ~do_skip_time_points )
                        
            for train_t = state_samples{test_s}
                
                mixed_feature                = train_feature;
                mixed_feature(:, train_t, :) = test_feature(:, test_t, :);
                
                log_likelihood(test_t, train_t) = 0;
                 
                %for t = 1:n_train_samples    
                for t = train_t:train_t

                    % Form log_P_x_c.
                    for c = 1:M
                        for i = 1:n_genes
                                                
                            for j = 1:num_exprs_per_gene
                                if (do_marginalize)
                                    if (t == train_t)    
                                        log_P_x_c(c, j, i, t) = (-0.5*sum(((mixed_feature(j, t, i) - u_c(c, test_s)).*(mixed_feature(j, t, i) - u_c(c, test_s)))./v_c(c, test_s)));
                                        log_P_x_c(c, j, i, t) = log_P_x_c(c, j, i, t) - log_product_stdev_c(c, test_s);
                                        log_P_x_c(c, j, i, t) = log_P_c(c, test_s) + log_P_x_c(c, j, i, t);
                                    else
                                        log_P_x_c(c, j, i, t) = 0; 
                                    end
                                else
                                    log_P_x_c(c, j, i, t) = (-0.5*sum(((mixed_feature(j, t, i) - u_c(c, test_s)).*(mixed_feature(j, t, i) - u_c(c, test_s)))./v_c(c, test_s)));
                                    log_P_x_c(c, j, i, t) = log_P_x_c(c, j, i, t) - log_product_stdev_c(c, test_s);

                                    if (do_gene_identifiability)
                                        log_P_x_c(c, j, i, t) = log_P_c_given_X_g(c, i, test_s) + log_P_x_c(c, j, i, t);
                                    else
                                        log_P_x_c(c, j, i, t) = log_P_c(c, test_s) + log_P_x_c(c, j, i, t);
                                    end
                                end
                            end
                        end
                    end
                    
                    % Form log_P_tilda_x_c:
                    % 1. Compute M_x.
                    % 2. Subtract M_x from log_P_x_c.
                    for i = 1:n_genes
                        for j = 1:num_exprs_per_gene
                            M_x(j, 1, i, t) = max(log_P_x_c(:, j, i, t));
                        end
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_log_P_x_c = log_P_x_c(:, :, i)
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        
                        for c = 1:M
                            log_P_tilda_x_c(c, :, i, t) = log_P_x_c(c, :, i, t) - M_x(:, 1, i, t)';
                            
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            %debug_M_x = M_x(:, 1, i)'
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        end
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %for j = 1:num_exprs_per_gene
                            %debug_log_P_tilda_x_c     = log_P_tilda_x_c(:, j, i)
                            %debug_max_log_P_tilda_x_c = max(log_P_tilda_x_c(:, j, i))
                            %pause;
                        %end
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    end
                    
                    % Compute P(c|x, i), P(c|X, i) and P(c|w, i).
                    for i = 1:n_genes
                        
                        % Compute P(c|x, i).
                        for c = 1:M
                             P_c_given_x_i(:, c, i, t) = (exp(log_P_tilda_x_c(c, :, i, t)))';
                        end
                        
                        % Normalize P(c|x, i).
                        for j = 1:num_exprs_per_gene
                            P_c_given_x_i(j, :, i, t) = P_c_given_x_i(j, :, i, t)./sum(P_c_given_x_i(j, :, i, t));
                            
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            %debug_P_c_given_x_i     = P_c_given_x_i(j, :, i)
                            %debug_sum_P_c_given_x_i = sum(P_c_given_x_i(j, :, i))
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        end
                        
                        % Compute P(c|X, i).
                        for c = 1:M
                            P_c_given_X_i(c, 1, i, t) = sum(P_c_given_x_i(:, c, i, t));
                        end
                        
                        % Normalize P(c|X, i).
                        P_c_given_X_i(:, 1, i, t) = P_c_given_X_i(:, 1, i, t)./sum(P_c_given_X_i(:, 1, i, t));
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_P_c_given_X_i     = P_c_given_X_i(:, 1, i)
                        %debug_sum_P_c_given_X_i = sum(P_c_given_X_i(:, 1, i))
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            
                        % Compute P(c|w, i).
                        for j = 1:num_tags_per_gene
                                % Consider only valid words.
                                if(word(j, 1, i) > 0)
                                    if (do_gene_identifiability)
                                        %P_c_given_w_i(j, :, i, t) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_given_w_g(j, :, i, test_s)';
                                        P_c_given_w_i(j, :, i, t) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_given_X_i(:, 1, i, t);
                                    else
                                        P_c_given_w_i(j, :, i, t) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_given_X_i(:, 1, i, t);
                                    end

                                    if (do_count_GO_tags_once)
                                        if (do_arithmetic_avg)
                                            if (do_gene_identifiability)
                                                log_P_w_given_X(j, 1, i, t) = safe_log((sum(P_c_given_w_i(j, :, i, t))/n_train_samples));
                                            else
                                                log_P_w_given_X(j, 1, i, t) = safe_log((sum(P_c_given_w_i(j, :, i, t))/n_train_samples));
                                            end
                                        else
                                            if (do_gene_identifiability)
                                                log_P_w_given_X(j, 1, i, t) = safe_log((sum(P_c_given_w_i(j, :, i, t)))^(1/n_train_samples));
                                            else
                                                log_P_w_given_X(j, 1, i, t) = safe_log((sum(P_c_given_w_i(j, :, i, t)))^(1/n_train_samples));
                                            end
                                        end
                                    else
                                        if (do_gene_identifiability)
                                            log_P_w_given_X(j, 1, i, t) = safe_log(sum(P_c_given_w_i(j, :, i, t)));
                                        else
                                            log_P_w_given_X(j, 1, i, t) = safe_log(sum(P_c_given_w_i(j, :, i, t)));
                                        end
                                    end
                                end
                        end
                        
                        % Normalize P(c|w, i).
                        for j = 1:num_tags_per_gene
                            % Consider only valid words.
                            if(word(j, 1, i) > 0)
                                P_c_given_w_i(j, :, i, t) = P_c_given_w_i(j, :, i, t)./sum(P_c_given_w_i(j, :, i, t));
                                
                                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                                %debug_P_c_given_w_i     = P_c_given_w_i(j, :, i)
                                %debug_sum_P_c_given_w_i = sum(P_c_given_w_i(j, :, i))
                                %pause;
                                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            end
                        end
                        
                    end                                    
                    
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Log likelihood computation.
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    
                    for i = 1:n_genes
                        
                        for j = 1:num_exprs_per_gene
                            log_P_tilda_x(j, 1, i, t) = safe_log(sum(exp(log_P_tilda_x_c(:, j, i, t)))); 
                            
                            log_likelihood(test_t, train_t) = log_likelihood(test_t, train_t) + M_x(j, 1, i, t);
                            log_likelihood(test_t, train_t) = log_likelihood(test_t, train_t) + log_P_tilda_x(j, 1, i, t);
                        end
                        
                        if (~do_arithmetic_avg)
                            for j = 1:num_tags_per_gene
                                % Consider only valid words.
                                if (word(j, 1, i) > 0)                                                                
                                    log_likelihood(test_t, train_t) = log_likelihood(test_t, train_t) + (log_P_w_given_X(j, 1, i, t)/n_go_tags_g(i));
                                end
                            end
                        end
                    end

                end

                if (do_arithmetic_avg)
                    for i = 1:n_genes
                        for j = 1:num_tags_per_gene
                            sum_over_t = 0;
                            for t = train_t:train_t
                                if (word(j, 1, i) > 0)
                                    sum_over_t = sum_over_t + exp(log_P_w_given_X(j, 1, i, t));
                                end
                            end
                            log_likelihood(test_t, train_t) = log_likelihood(test_t, train_t) + (safe_log(sum_over_t)/n_go_tags_g(i));
                        end
                    end
                end

            end

        else

            log_P_s(test_s) = 0;

            for t = test_t:test_t

                    % Form log_P_x_c_s.
                    for c = 1:M
                        for i = 1:n_genes
                                                
                            for j = 1:num_exprs_per_gene
                                log_P_x_c_s(c, j, i) = (-0.5*sum(((test_feature(j, t, i) - u_c(c, test_s)).*(test_feature(j, t, i) - u_c(c, test_s)))./v_c(c, test_s)));
                                log_P_x_c_s(c, j, i) = log_P_x_c_s(c, j, i) - log_product_stdev_c(c, test_s);

                                if (do_gene_identifiability)
                                    log_P_x_c_s(c, j, i) = log_P_c_given_X_g(c, i, test_s) + log_P_x_c_s(c, j, i);
                                else
                                    log_P_x_c_s(c, j, i) = log_P_c(c, test_s) + log_P_x_c_s(c, j, i);
                                end
                            end
                        end
                    end
                    
                    % Form log_P_tilda_x_c_s:
                    % 1. Compute M_x_s.
                    % 2. Subtract M_x_s from log_P_x_c_s.
                    for i = 1:n_genes
                        for j = 1:num_exprs_per_gene
                            M_x_s(j, 1, i) = max(log_P_x_c_s(:, j, i));
                        end
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_log_P_x_c_s = log_P_x_c_s(:, :, i)
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        
                        for c = 1:M
                            log_P_tilda_x_c_s(c, :, i) = log_P_x_c_s(c, :, i) - M_x_s(:, 1, i)';
                            
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            %debug_M_x_s = M_x_s(:, 1, i)'
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        end
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %for j = 1:num_exprs_per_gene
                            %debug_log_P_tilda_x_c_s     = log_P_tilda_x_c_s(:, j, i)
                            %debug_max_log_P_tilda_x_c_s = max(log_P_tilda_x_c_s(:, j, i))
                            %pause;
                        %end
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                    end
                    
                    % Compute P(c|x, i), P(c|X, i) and P(c|w, i).
                    for i = 1:n_genes
                        
                        % Compute P(c|x, i).
                        for c = 1:M
                             P_c_s_given_x_i(:, c, i) = (exp(log_P_tilda_x_c_s(c, :, i)))';
                        end
                        
                        % Normalize P(c|x, i).
                        for j = 1:num_exprs_per_gene
                            P_c_s_given_x_i(j, :, i) = P_c_s_given_x_i(j, :, i)./sum(P_c_s_given_x_i(j, :, i));
                            
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            %debug_P_c_s_given_x_i     = P_c_s_given_x_i(j, :, i)
                            %debug_sum_P_c_s_given_x_i = sum(P_c_s_given_x_i(j, :, i))
                            %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        end
                        
                        % Compute P(c|X, i).
                        for c = 1:M
                            P_c_s_given_X_i(c, 1, i) = sum(P_c_s_given_x_i(:, c, i));
                        end
                        
                        % Normalize P(c|X, i).
                        P_c_s_given_X_i(:, 1, i) = P_c_s_given_X_i(:, 1, i)./sum(P_c_s_given_X_i(:, 1, i));
                        
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                        %debug_P_c_s_given_X_i     = P_c_s_given_X_i(:, 1, i)
                        %debug_sum_P_c_s_given_X_i = sum(P_c_s_given_X_i(:, 1, i))
                        %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            
                        % Compute P(c|w, i).
                        for j = 1:num_tags_per_gene
                                % Consider only valid words.
                                if(word(j, 1, i) > 0)
                                    if (do_gene_identifiability)
                                        %P_c_s_given_w_i(j, :, i) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_given_w_g(j, :, i, test_s)';
                                        P_c_s_given_w_i(j, :, i) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_s_given_X_i(:, 1, i);
                                    else
                                        P_c_s_given_w_i(j, :, i) = P_w_given_c(:, word(j, 1, i), test_s).*P_c_s_given_X_i(:, 1, i);
                                    end

                                    if (do_count_GO_tags_once)
                                        if (do_arithmetic_avg)
                                            log_P_w_given_X_s(j, 1, i) = safe_log(sum(P_c_s_given_w_i(j, :, i)));
                                        else
                                            log_P_w_given_X_s(j, 1, i) = safe_log(sum(P_c_s_given_w_i(j, :, i)));
                                        end
                                    else
                                        log_P_w_given_X_s(j, 1, i) = safe_log(sum(P_c_s_given_w_i(j, :, i)));
                                    end                                
                                end
                        end
                        
                        % Normalize P(c|w, i).
                        for j = 1:num_tags_per_gene
                            % Consider only valid words.
                            if(word(j, 1, i) > 0)
                                P_c_s_given_w_i(j, :, i) = P_c_s_given_w_i(j, :, i)./sum(P_c_s_given_w_i(j, :, i));
                                
                                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                                %debug_P_c_s_given_w_i     = P_c_s_given_w_i(j, :, i)
                                %debug_sum_P_c_s_given_w_i = sum(P_c_s_given_w_i(j, :, i))
                                %pause;
                                %%%%%%%%%%%% Debugging%%%%%%%%%%%
                            end
                        end
                        
                    end                                    
                    
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Log likelihood computation.
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    
                    for i = 1:n_genes
                        
                        for j = 1:num_exprs_per_gene
                            log_P_tilda_x_s(j, 1, i) = safe_log(sum(exp(log_P_tilda_x_c_s(:, j, i)))); 
                            
                            log_P_s(test_s) = log_P_s(test_s) + M_x_s(j, 1, i);
                            log_P_s(test_s) = log_P_s(test_s) + log_P_tilda_x_s(j, 1, i);
                        end
                        
                        if (~do_arithmetic_avg)
                            for j = 1:num_tags_per_gene
                                % Consider only valid words.
                                if (word(j, 1, i) > 0)
                                    log_P_s(test_s) = log_P_s(test_s) + (log_P_w_given_X_s(j, 1, i)/n_go_tags_g(i));
                                end
                            end
                        end
                    end

            end

            if (do_arithmetic_avg)
                for i = 1:n_genes
                    for j = 1:num_tags_per_gene
                        sum_over_t = 0;
                        if (word(j, 1, i) > 0)
                            sum_over_t = sum_over_t + exp(log_P_w_given_X_s(j, 1, i));
                        end
                        log_P_s(test_s) = log_P_s(test_s) + (safe_log(sum_over_t)/n_go_tags_g(i));
                    end
                end
            end

        end

    end

    [state_pred_probs(test_t, :), log_sum_exp_p_s] = scale_by_sum_log_vec(log_P_s);
end

%Debugging.
log_likelihood
state_pred_probs

if ( ~do_skip_time_points )

    % Convert log likelihoods to prediction probabilities.
    for test_t = 1:n_test_samples
        M_t(test_t)                     = max(log_likelihood(test_t, :));
        log_tilda_likelihood(test_t, :) = log_likelihood(test_t, :) - M_t(test_t);
    end

    sum_over_t = zeros(n_test_samples, 1);
    for test_t = 1:n_test_samples
        time_pred_probs(test_t, :) = exp(log_tilda_likelihood(test_t, :));
        sum_over_t(test_t)         = sum(time_pred_probs(test_t, :));
        time_pred_probs(test_t, :) = time_pred_probs(test_t, :) ./ sum_over_t(test_t);
    end

end
